from __future__ import annotations

from collections import defaultdict
from collections.abc import Collection, Iterator, Mapping
from dataclasses import dataclass
from enum import Enum

from networkx import DiGraph, ancestors, descendants

from synthesize.config import After, ResolvedFlow, ResolvedNode


@dataclass(frozen=True)
class FlowState:
    graph: DiGraph
    flow: ResolvedFlow
    statuses: dict[str, Status]

    @classmethod
    def from_flow(cls, flow: ResolvedFlow) -> FlowState:
        graph = DiGraph()

        for id, node in flow.nodes.items():
            graph.add_node(id)
            for t in node.triggers:
                if isinstance(t, After):
                    for predecessor_id in t.after:
                        graph.add_edge(predecessor_id, id)

        return FlowState(
            graph=graph,
            flow=flow,
            statuses={id: Status.Pending for id in graph.nodes},
        )

    def nodes_by_status(self) -> Mapping[Status, Collection[ResolvedNode]]:
        d = defaultdict(list)
        for id, s in self.statuses.items():
            d[s].append(self.flow.nodes[id])
        return d

    def ready_nodes(self) -> Collection[ResolvedNode]:
        return tuple(
            self.flow.nodes[id]
            for id in self.graph.nodes
            if self.statuses[id] is Status.Pending
            and all(
                self.statuses[a]
                in (
                    Status.Succeeded,
                    Status.Waiting,
                )
                for a in ancestors(self.graph, id)
            )
        )

    def mark_success(self, *nodes: ResolvedNode) -> None:
        self.mark(*nodes, status=Status.Succeeded)

    def mark_failure(self, *nodes: ResolvedNode) -> None:
        self.mark(*nodes, status=Status.Failed)

    def mark_pending(self, *nodes: ResolvedNode) -> None:
        self.mark(*nodes, status=Status.Pending)

    def mark_running(self, *nodes: ResolvedNode) -> None:
        self.mark(*nodes, status=Status.Running)

    def mark(self, *nodes: ResolvedNode, status: Status) -> None:
        for node in nodes:
            self.statuses[node.id] = status

    def children(self, node: ResolvedNode) -> Collection[ResolvedNode]:
        return tuple(self.flow.nodes[id] for id in self.graph.successors(node.id))

    def descendants(self, node: ResolvedNode) -> Collection[ResolvedNode]:
        return tuple(self.flow.nodes[id] for id in descendants(self.graph, node.id))

    def all_done(self) -> bool:
        return all(status is Status.Succeeded for status in self.statuses.values())

    def nodes(self) -> Iterator[ResolvedNode]:
        yield from self.flow.nodes.values()


class Status(Enum):
    Pending = "pending"
    Waiting = "waiting"
    Starting = "starting"
    Running = "running"
    Succeeded = "succeeded"
    Failed = "failed"



from __future__ import annotations

import asyncio
from pathlib import Path
from time import monotonic
from typing import Optional

import typer.rich_utils as ru
from click.exceptions import Exit
from pydantic import ValidationError
from rich.console import Console
from rich.json import JSON
from rich.panel import Panel
from rich.style import Style
from rich.text import Text
from typer import Argument, Option, Typer

from synthesize.config import Config
from synthesize.orchestrator import Orchestrator

ru.STYLE_HELPTEXT = ""

cli = Typer(pretty_exceptions_enable=False)


@cli.command()
def run(
    flow: str = Argument(
        default="default",
        help="The flow to execute.",
    ),
    config: Optional[Path] = Option(
        default=None,
        exists=True,
        readable=True,
        show_default=True,
        envvar="SYNTHFILE",
        help="The path to the configuration file to execute.",
    ),
    mermaid: bool = Option(
        default=False,
        help="If enabled, output a description of the flow as a Mermaid diagram, and don't run the flow.",
    ),
    dry: bool = Option(
        default=False,
        help="If enabled, do not run actually run the flow.",
    ),
) -> None:
    start_time = monotonic()

    console = Console()

    config = config or find_config_file(console)

    try:
        parsed_config = Config.from_file(config)
    except ValidationError as e:
        for err in e.errors():
            loc = ".".join(map(str, err["loc"]))
            msg = err["msg"]
            console.print(f"[red]ERROR[/red] {loc} -> {msg}")
        raise Exit(code=1)

    if dry:
        console.print(
            Panel(
                JSON(parsed_config.model_dump_json(exclude_unset=True)),
                title="Configuration",
                title_align="left",
            )
        )

    resolved = parsed_config.resolve()

    try:
        selected_flow = resolved[flow]
    except KeyError:
        sep = "\n  "
        available_flows = sep + sep.join(resolved.keys())
        console.print(
            Text(
                f"No flow named '{flow}'. Available flows:{available_flows}",
                style=Style(color="red"),
            )
        )
        raise Exit(code=1)

    if mermaid:
        print(selected_flow.mermaid())
        return

    if dry:
        return

    controller = Orchestrator(flow=selected_flow, console=console)

    try:
        asyncio.run(controller.run())
    except KeyboardInterrupt:
        raise Exit(code=0)
    finally:
        end_time = monotonic()

        console.print(Text(f"Finished in {end_time - start_time:.3f} seconds."))


def find_config_file(console: Console) -> Path:
    cwd = Path.cwd()
    for dir in (cwd, *cwd.parents):
        contents = set(dir.iterdir())
        for name in ("synth.yaml",):
            if (path := dir / name) in contents:
                return path

        if dir / ".git" in contents:
            break

    console.print(Text("Failed to find a Synthesize config file", style=Style(color="red")))
    raise Exit(code=1)



from __future__ import annotations

import hashlib
from asyncio import Task, create_task, sleep
from typing import Awaitable, Callable, Optional, TypeVar

T = TypeVar("T")


def delay(delay: float, fn: Callable[[], Awaitable[T]], name: Optional[str] = None) -> Task[T]:
    async def delayed() -> T:
        await sleep(delay)
        return await fn()

    return create_task(delayed(), name=name)


def md5(data: bytes | str) -> str:
    return hashlib.md5(data if isinstance(data, bytes) else data.encode()).hexdigest()



from __future__ import annotations

import signal
from asyncio import Queue, Task, create_task, gather, get_running_loop, sleep
from collections.abc import Iterable
from pathlib import Path
from tempfile import TemporaryDirectory

from rich.console import Console
from watchfiles import awatch

from synthesize.config import ResolvedFlow, ResolvedNode, Restart, Watch
from synthesize.execution import Execution
from synthesize.messages import (
    ExecutionCompleted,
    ExecutionStarted,
    Heartbeat,
    Message,
    Quit,
    WatchPathChanged,
)
from synthesize.renderer import Renderer
from synthesize.state import FlowState, Status


class Orchestrator:
    def __init__(self, flow: ResolvedFlow, console: Console):
        self.flow = flow
        self.console = console

        self.state = FlowState.from_flow(flow=flow)
        self.renderer = Renderer(state=self.state, console=console)

        self.inbox: Queue[Message] = Queue()

        self.executions: dict[str, Execution] = {}
        self.waiters: dict[str, Task[Execution]] = {}
        self.watchers: dict[str, Task[None]] = {}
        self.heartbeat: Task[None] | None = None

    async def run(self) -> None:
        if not self.flow.nodes:
            return

        with TemporaryDirectory(prefix="synth-") as tmpdir, self.renderer:
            tmp_dir = Path(tmpdir)

            try:
                await self.start_heartbeat()
                await self.start_watchers()
                await self.start_ready_targets(tmp_dir=tmp_dir)

                await self.handle_messages(tmp_dir=tmp_dir)
            finally:
                self.renderer.handle_shutdown_start()

                if self.heartbeat is not None:
                    self.heartbeat.cancel()

                for watcher in self.watchers.values():
                    watcher.cancel()

                await gather(*self.watchers.values(), return_exceptions=True)

                for execution in self.executions.values():
                    execution.terminate()

                await gather(*(e.wait() for e in self.executions.values()), return_exceptions=True)

                self.renderer.handle_shutdown_end()

    async def handle_messages(self, tmp_dir: Path) -> None:
        signal.signal(signal.SIGINT, lambda sig, frame: self.inbox.put_nowait(Quit()))

        while True:
            match message := await self.inbox.get():
                case ExecutionStarted(node=node):
                    self.state.mark_running(node)

                case ExecutionCompleted(node=node, exit_code=exit_code):
                    if self.state.statuses[node.id] is not Status.Pending:
                        for t in node.triggers:
                            if isinstance(t, Restart):
                                if self.state.statuses[node.id] is not Status.Waiting:
                                    self.state.mark(node, status=Status.Waiting)

                                    def waiting_to_pending() -> None:
                                        if self.state.statuses[node.id] is Status.Waiting:
                                            self.state.mark_pending(node)

                                    get_running_loop().call_later(t.delay, waiting_to_pending)
                                break
                        else:
                            if exit_code == 0:
                                self.state.mark_success(node)
                            else:
                                self.state.mark_failure(node)

                    self.state.mark_pending(*self.state.children(node))

                case WatchPathChanged(node=node):
                    if e := self.executions.get(node.id):
                        e.terminate()
                        self.state.mark_pending(node)

                case Quit():
                    return

            await self.start_ready_targets(tmp_dir=tmp_dir)

            self.renderer.handle_message(message)

            if self.state.all_done() and not self.watchers:
                return

    async def start_heartbeat(self) -> None:
        async def heartbeat() -> None:
            while True:
                await sleep(1 / 10)
                await self.inbox.put(Heartbeat())

        self.heartbeat = create_task(heartbeat())

    async def start_ready_targets(self, tmp_dir: Path) -> None:
        for node in self.state.ready_nodes():
            if e := self.executions.get(node.id):
                if not e.has_exited:
                    continue

            self.state.mark(node, status=Status.Waiting)

            e = await Execution.start(
                node=node,
                args=self.flow.args,
                envs=self.flow.envs,
                tmp_dir=tmp_dir,
                width=self.console.width - self.renderer.prefix_width,
                events=self.inbox,
            )
            self.executions[node.id] = e
            self.waiters[node.id] = create_task(e.wait())

    async def start_watchers(self) -> None:
        for node in self.flow.nodes.values():
            for trigger in node.triggers:
                if isinstance(trigger, Watch):
                    self.watchers[node.id] = create_task(
                        watch(
                            node=node,
                            paths=trigger.watch,
                            events=self.inbox,
                        )
                    )


async def watch(node: ResolvedNode, paths: Iterable[str | Path], events: Queue[Message]) -> None:
    async for changes in awatch(*paths):
        await events.put(WatchPathChanged(node=node, changes=changes))



from __future__ import annotations

from typing import ClassVar, Type, TypeVar

import yaml
from pydantic import BaseModel, ConfigDict

C = TypeVar("C", bound="Model")


class Model(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(
        frozen=True,
        use_enum_values=True,
        extra="forbid",
        coerce_numbers_to_str=True,
    )

    @classmethod
    def model_validate_yaml(cls: Type[C], y: str) -> C:
        return cls.model_validate(yaml.safe_load(y))



from __future__ import annotations

from datetime import datetime
from functools import cached_property
from types import TracebackType
from typing import Type

from rich.console import Console, Group, RenderableType
from rich.live import Live
from rich.rule import Rule
from rich.style import Style
from rich.table import Table
from rich.text import Text
from typing_extensions import assert_never
from watchfiles import Change

from synthesize.messages import (
    ExecutionCompleted,
    ExecutionOutput,
    ExecutionStarted,
    Message,
    WatchPathChanged,
)
from synthesize.state import FlowState, Status

prefix_format = "{timestamp:%H:%M:%S} {id}  "
internal_format = "{timestamp:%H:%M:%S}"
CHANGE_TO_STYLE = {
    Change.added: Style(color="green"),
    Change.deleted: Style(color="red"),
    Change.modified: Style(color="yellow"),
}


class Renderer:
    def __init__(self, state: FlowState, console: Console):
        self.state = state
        self.console = console

        self.live = Live(console=console, auto_refresh=False)

    def __enter__(self) -> None:
        self.live.start(refresh=True)

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.live.stop()

    def handle_message(self, message: Message) -> None:
        match message:
            case ExecutionOutput() as msg:
                self.handle_command_message(msg)

            case ExecutionStarted() | ExecutionCompleted() | WatchPathChanged() as msg:
                self.handle_lifecycle_message(msg)

        self.update(message)

    def info(self, event: Message) -> RenderableType:
        table = Table.grid(padding=(1, 1, 0, 0), expand=False)

        status_table = Table.grid(padding=(2, 2, 0, 0), expand=False)

        nodes_by_status = self.state.nodes_by_status()
        node_status_displays = []
        for status in (
            Status.Running,
            Status.Waiting,
            Status.Pending,
            Status.Succeeded,
            Status.Failed,
        ):
            nodes_with_status = nodes_by_status[status]
            if nodes_with_status:
                node_status_displays.append(
                    Text.assemble(
                        status.value.capitalize(),
                        " ",
                        Text(" ").join(
                            Text(t.id, style=Style(color="black", bgcolor=t.color))
                            for t in sorted(nodes_with_status, key=lambda t: t.id)
                        ),
                    )
                )

        status_table.add_row(*node_status_displays)

        table.add_row(
            internal_format.format_map({"timestamp": event.timestamp}),
            status_table,
        )

        return Group(
            Rule(style=(Style(color="green" if nodes_by_status[Status.Running] else "yellow"))),
            table,
        )

    def render_prefix(
        self, message: ExecutionOutput | ExecutionStarted | ExecutionCompleted | WatchPathChanged
    ) -> str:
        return prefix_format.format_map(
            {"id": message.node.id, "timestamp": message.timestamp}
        ).ljust(self.prefix_width)

    def handle_command_message(self, message: ExecutionOutput) -> None:
        prefix = Text(
            self.render_prefix(message),
            style=Style(color=message.node.color),
        )

        body = Text.from_ansi(message.text)

        g = Table.grid()
        g.add_row(prefix, body)

        self.console.print(g)

    def handle_lifecycle_message(
        self, message: ExecutionStarted | ExecutionCompleted | WatchPathChanged
    ) -> None:
        prefix = Text.from_markup(
            self.render_prefix(message),
            style=Style(color=message.node.color, dim=True),
        )

        parts: tuple[str | tuple[str, str] | tuple[str, Style] | Text, ...]

        match message:
            case ExecutionStarted(node=node, pid=pid):
                parts = (
                    (node.id, node.color),
                    f" started (pid {pid})",
                )
            case ExecutionCompleted(node=node, pid=pid, exit_code=exit_code, duration=duration):
                parts = (
                    (node.id, node.color),
                    f" (pid {pid}) exited with code ",
                    (str(exit_code), "green" if exit_code == 0 else "red"),
                    f" in {duration.total_seconds() :.3f} seconds",
                )
            case WatchPathChanged(node=node):
                changes = Text(" ").join(
                    Text(path, style=CHANGE_TO_STYLE[change]) for change, path in message.changes
                )

                parts = (
                    "Running ",
                    (node.id, node.color),
                    " due to detected changes: ",
                    changes,
                )
            case _:
                assert_never(message)

        body = Text.assemble(
            *parts,
            style=Style(dim=True),
        )

        g = Table.grid()
        g.add_row(prefix, body)

        self.console.print(g)

    def handle_shutdown_start(self) -> None:
        self.live.update(Group(Rule(), Text("Shutting down...")), refresh=True)

    def handle_shutdown_end(self) -> None:
        self.live.update(Rule(), refresh=True)

    def update(self, message: Message) -> None:
        self.live.update(self.info(message), refresh=True)

    @cached_property
    def prefix_width(self) -> int:
        now = datetime.now()

        return len(
            max(
                (
                    prefix_format.format_map({"timestamp": now, "id": t.id})
                    for t in self.state.nodes()
                ),
                key=len,
            )
        )



from __future__ import annotations

import os
from asyncio import Queue, Task, create_task
from asyncio.subprocess import PIPE, STDOUT, Process, create_subprocess_exec
from dataclasses import dataclass, field
from datetime import timedelta
from pathlib import Path
from signal import SIGKILL, SIGTERM
from stat import S_IEXEC
from time import monotonic

from synthesize.config import Args, Envs, ResolvedNode
from synthesize.messages import ExecutionCompleted, ExecutionOutput, ExecutionStarted, Message


def write_script(node: ResolvedNode, args: Args, tmp_dir: Path) -> Path:
    path = tmp_dir / f"{node.id}-{node.uid}"

    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(
        node.target.render(
            args=args
            | node.target.args
            | node.args
            | {
                "id": node.id,
            }
        )
    )
    path.chmod(path.stat().st_mode | S_IEXEC)

    return path


@dataclass(frozen=True)
class Execution:
    node: ResolvedNode

    events: Queue[Message] = field(repr=False)

    process: Process
    start_time: float
    reader: Task[None]

    @classmethod
    async def start(
        cls,
        node: ResolvedNode,
        args: Args,
        envs: Envs,
        tmp_dir: Path,
        width: int,
        events: Queue[Message],
    ) -> Execution:
        path = write_script(node=node, args=args, tmp_dir=tmp_dir)

        start_time = monotonic()

        process = await create_subprocess_exec(
            program=path,
            stdout=PIPE,
            stderr=STDOUT,
            env=os.environ
            | envs
            | node.target.envs
            | node.envs
            | {
                "FORCE_COLOR": "1",
                "COLUMNS": str(width),
            }
            | {
                "SYNTH_NODE_ID": node.id,
            },
            preexec_fn=os.setsid,
        )

        reader = create_task(
            read_output(
                node=node,
                process=process,
                events=events,
            ),
            name=f"Read output for {node.id}",
        )

        await events.put(ExecutionStarted(node=node, pid=process.pid))

        return cls(
            node=node,
            events=events,
            process=process,
            start_time=start_time,
            reader=reader,
        )

    @property
    def pid(self) -> int:
        return self.process.pid

    @property
    def exit_code(self) -> int | None:
        return self.process.returncode

    @property
    def has_exited(self) -> bool:
        return self.exit_code is not None

    def _send_signal(self, signal: int) -> None:
        if self.has_exited:
            return None

        try:
            os.killpg(os.getpgid(self.process.pid), signal)
        except ProcessLookupError:
            # process exited before we could send the signal
            pass

    def terminate(self) -> None:
        self._send_signal(SIGTERM)

    def kill(self) -> None:
        self._send_signal(SIGKILL)

    async def wait(self) -> Execution:
        exit_code = await self.process.wait()
        end_time = monotonic()

        await self.reader

        await self.events.put(
            ExecutionCompleted(
                node=self.node,
                pid=self.pid,
                exit_code=exit_code,
                duration=timedelta(seconds=end_time - self.start_time),
            )
        )

        return self


async def read_output(node: ResolvedNode, process: Process, events: Queue[Message]) -> None:
    if process.stdout is None:  # pragma: unreachable
        raise Exception(f"{process} does not have an associated stream reader")

    while True:
        line = await process.stdout.readline()
        if not line:
            break

        await events.put(
            ExecutionOutput(
                node=node,
                text=line.decode("utf-8").rstrip(),
            )
        )


# need to track which trigger caused the node to run,
# because that changes the semantics of the restart
# the manager should protect itself from multiple restarts?



from datetime import datetime, timedelta

from pydantic import Field
from watchfiles import Change

from synthesize.config import ResolvedNode
from synthesize.model import Model


class Message(Model):
    timestamp: datetime = Field(default_factory=datetime.now)


class ExecutionStarted(Message):
    node: ResolvedNode
    pid: int


class ExecutionCompleted(Message):
    node: ResolvedNode
    pid: int
    exit_code: int
    duration: timedelta


class ExecutionOutput(Message):
    node: ResolvedNode
    text: str


class WatchPathChanged(Message):
    node: ResolvedNode
    changes: set[tuple[Change, str]]


class Heartbeat(Message):
    pass


class Quit(Message):
    pass



from __future__ import annotations

import shlex
import shutil
from collections.abc import Mapping
from colorsys import hsv_to_rgb
from functools import cached_property
from pathlib import Path
from random import random
from textwrap import dedent
from typing import Annotated, Union

from identify.identify import tags_from_path
from jinja2 import Environment
from networkx import DiGraph
from pydantic import Field, field_validator
from rich.color import Color
from typing_extensions import assert_never

from synthesize.model import Model
from synthesize.utils import md5

Args = dict[
    Annotated[
        str,
        Field(
            # https://jinja.palletsprojects.com/en/3.1.x/api/#notes-on-identifiers
            pattern=r"[a-zA-Z_][a-zA-Z0-9_]*(\.[a-zA-Z_][a-zA-Z0-9_]*)*",
            min_length=1,
        ),
    ],
    object,
]
Envs = dict[Annotated[str, Field(min_length=1)], str]
ID = Annotated[str, Field(pattern=r"\w+")]


def random_color() -> str:
    triplet = Color.from_rgb(*(x * 255 for x in hsv_to_rgb(random(), 1, 0.7))).triplet

    if triplet is None:  # pragma: unreachable
        raise Exception("Failed to generate random color; please try again.")

    return triplet.hex


template_environment = Environment()


class Target(Model):
    commands: Annotated[str, Field(description="The commands to run for this target.")] = ""
    args: Annotated[
        Args,
        Field(
            description="Template arguments to apply to this target by default.",
        ),
    ] = {}
    envs: Annotated[
        Envs,
        Field(
            description="Environment variables to apply to this target by default.",
        ),
    ] = {}

    executable: Annotated[str, Field(description="The executable to run this target with.")] = (
        "sh -eu"
    )

    @field_validator("commands")
    @classmethod
    def dedent_commands(cls, commands: str) -> str:
        return dedent(commands).strip()

    def render(self, args: Args) -> str:
        exe, *exe_args = shlex.split(self.executable)
        which_exe = shutil.which(exe)
        if which_exe is None:
            raise Exception(f"Failed to find absolute path to executable for {exe}")

        template = template_environment.from_string(
            "\n".join(
                (
                    f"#!{shlex.join((which_exe, *exe_args))}",
                    "",
                    self.commands,
                )
            )
        )

        return template.render(args)


class Once(Model):
    pass


class After(Model):
    after: Annotated[
        tuple[str, ...],
        Field(
            min_length=1,
            description="The IDs of the nodes to wait for.",
        ),
    ]


class Restart(Model):
    delay: Annotated[
        float,
        Field(
            description="The delay before restarting the command after it exits.",
            ge=0,
        ),
    ] = 1


class Watch(Model):
    watch: Annotated[
        tuple[str, ...],
        Field(
            description="The paths to watch for changes. Directories are watched recursively.",
        ),
    ]


AnyTrigger = Union[
    Once,
    After,
    Restart,
    Watch,
]


class ResolvedNode(Model):
    id: str

    target: Target
    args: Annotated[
        Args,
        Field(
            description="Template arguments to apply to this node.",
        ),
    ] = {}
    envs: Annotated[
        Envs,
        Field(
            description="Environment variables to apply to this node.",
        ),
    ] = {}

    triggers: tuple[AnyTrigger, ...] = (Once(),)

    color: Annotated[str, Field(default_factory=random_color)]

    @cached_property
    def uid(self) -> str:
        return md5(self.model_dump_json(exclude={"color"}).encode())


class Node(Model):
    target: Annotated[
        Target | ID,
        Field(
            description="The target to run for this node. It may either be the name of a pre-defined target, or a full target definition.",
        ),
    ]
    args: Annotated[
        Args,
        Field(
            description="Template arguments to apply to this node.",
        ),
    ] = {}
    envs: Annotated[
        Envs,
        Field(
            description="Environment variables to apply to this node.",
        ),
    ] = {}

    triggers: Annotated[
        tuple[AnyTrigger | ID, ...],
        Field(
            description="The list of triggers for this node. Each trigger may be the name of a pre-defined trigger, or a full trigger definition.",
        ),
    ] = (Once(),)

    color: Annotated[
        str,
        Field(
            default_factory=random_color,
            description="The color that will be used to help differentiate this node from others.",
        ),
    ]

    def resolve(
        self,
        id: str,
        targets: Mapping[str, Target],
        triggers: Mapping[str, AnyTrigger],
    ) -> ResolvedNode:
        return ResolvedNode(
            id=id,
            target=targets[self.target] if isinstance(self.target, str) else self.target,
            args=self.args,
            envs=self.envs,
            triggers=tuple(triggers[t] if isinstance(t, str) else t for t in self.triggers),
            color=self.color,
        )


class ResolvedFlow(Model):
    nodes: dict[ID, ResolvedNode]
    args: Annotated[
        Args,
        Field(
            description="Template arguments to apply to all nodes in this flow.",
        ),
    ] = {}
    envs: Annotated[
        Envs,
        Field(
            description="Environment variables to apply to all nodes in this flow.",
        ),
    ] = {}

    @cached_property
    def graph(self) -> DiGraph:
        graph = DiGraph()

        for id, node in self.nodes.items():
            graph.add_node(id)
            for t in node.triggers:
                if isinstance(t, After):
                    for predecessor_id in t.after:
                        graph.add_edge(predecessor_id, id)

        return graph

    def mermaid(self) -> str:
        lines = ["flowchart TD"]

        seen_watches = set()
        for id, node in self.nodes.items():
            lines.append(f"{node.id}({id})")

            for t in node.triggers:
                match t:
                    case Once():
                        pass
                    case After(after=after):
                        for a in after:
                            lines.append(f"{self.nodes[a].id} --> {node.id}")
                    case Restart(delay=delay):
                        lines.append(f"{node.id} -->|∞ {delay:.3g}s| {node.id}")
                    case Watch(watch=paths):
                        text = "\n".join(paths)
                        h = md5("".join(paths))
                        if h not in seen_watches:
                            seen_watches.add(h)
                            lines.append(f'w_{h}[("{text}")]')
                        lines.append(f"w_{h} -->|👁| {node.id}")
                    case never:
                        assert_never(never)

        return "\n  ".join(lines).strip()


class Flow(Model):
    nodes: Annotated[
        Mapping[ID, Node],
        Field(
            description="Mapping of IDs to nodes.",
        ),
    ] = {}
    args: Annotated[
        Args,
        Field(
            description="Template arguments to apply to all nodes in this flow.",
        ),
    ] = {}
    envs: Annotated[
        Envs,
        Field(
            description="Environment variables to apply to all nodes in this flow.",
        ),
    ] = {}

    def resolve(
        self,
        targets: Mapping[ID, Target],
        triggers: Mapping[ID, AnyTrigger],
    ) -> ResolvedFlow:
        return ResolvedFlow(
            nodes={id: node.resolve(id, targets, triggers) for id, node in self.nodes.items()},
            args=self.args,
            envs=self.envs,
        )


class Config(Model):
    flows: Annotated[
        Mapping[ID, Flow],
        Field(
            description="A mapping of IDs to flows.",
        ),
    ] = {}
    targets: Annotated[
        Mapping[ID, Target],
        Field(
            description="A mapping of IDs to targets.",
        ),
    ] = {}
    triggers: Annotated[
        Mapping[ID, AnyTrigger],
        Field(
            description="A mapping of IDs to triggers.",
        ),
    ] = {}

    @classmethod
    def from_file(cls, file: Path) -> Config:
        tags = tags_from_path(str(file))

        if "yaml" in tags:
            return cls.model_validate_yaml(file.read_text())
        else:
            raise NotImplementedError("Currently, only YAML files are supported.")

    def resolve(self) -> Mapping[ID, ResolvedFlow]:
        return {id: flow.resolve(self.targets, self.triggers) for id, flow in self.flows.items()}
